
// Baseball simulation
// Source: Simulations of baseball pitches
//         (http://farside.ph.utexas.edu/teaching/329/)
// Simple (not optimized) implementation of Runge-Kutta 4
// Needs linking with math library (GCC option -lm)
// Public domain. Author: Petr Peringer, FIT TU Brno, 2021

#include <stdio.h>
#include <math.h>               // sin(), cos()

// Parameters:
const double g = 9.81;          // gravity acceleration [m/s^{-2}]
const double len = 18.44;       // distance [m]
double v0 = 85*1602/3600.0;               // initial speed [m/s]
double fi = 0;                  // angle of spin axis? [rad]
double theta = (1/360.0)*2*M_PI;          // initial angle of elevation [rad]
double omega = (1800/60.0)*2*M_PI;        // angular velocity [rad/s]
double B=4.1e-4;                // ? Magnus force coefficient

double D(double vx, double vy, double vz);      // drag coefficient function?

// System: baseball
#define SIZE 6                  // state vector size of the system
// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    const double vx = y[1];
    const double vy = y[3];
    const double vz = y[5];
    const double drag = -D(vx,vy,vz);
    // x-axis
    yin[0] = vx;      // dx/dt
    yin[1] = drag*vx + B*omega*(vz*sin(fi)-vy*cos(fi));
    // y-axis
    yin[2] = vy;      // dy/dt
    yin[3] = drag*vy + B*omega*vx*cos(fi);
    // z-axis
    yin[4] = vz;      // dz/dt
    yin[5] = -g - drag*vz - B*omega*vx*sin(fi);
}

// drag computation function
double D(double vx, double vy, double vz) {
    const double vd = 35; // m/s
    const double delta = 5; // m/s
    double v = sqrt(vx*vx+vy*vy+vz*vz);
    double fv = 0.0039 + 0.0058 / (1.0+exp((v-vd)/delta));
    return fv*v;
}

void RK4step(double t, double h, double *y);

//////////////////////////////////////////////////////////////////
// Experiment control
int main()
{
    double y[SIZE] = { 0, v0*cos(theta), 0, 0, 0, v0*sin(theta) };
    double stepsize = 1e-4;     // may be changed

    double t = 0;
    while (y[0] < len) {        // stop approximately at target x position
        printf("%8f % -8f % -8f  % -8f\n", t, y[0], y[2], y[4]);
        RK4step(t, stepsize, y);        // make step
        t += stepsize;          // advance the simulation time
    }
    printf("%8f % -8f % -8f  % -8f\n", t, y[0], y[2], y[4]);
    return 0;
}

//////////////////////////////////////////////////////////////////
// Runge-Kutta 4-th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ystart[SIZE];        // initial state
    // 4 stages results:
    double k1[SIZE];
    double k2[SIZE];
    double k3[SIZE];
    double k4[SIZE];

    int i;
    for (i = 0; i < SIZE; i++)  // save initial value
        ystart[i] = y[i];

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        k1[i] = h * yin[i];     // k1 = h * f(t, y(t))
        y[i] = ystart[i] + k1[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        k2[i] = h * yin[i];     // k2 = h * f(t+h/2, y(t)+k1/2)
        y[i] = ystart[i] + k2[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        k3[i] = h * yin[i];     // k3 = h * f(t+h/2, y(t)+k2/2)
        y[i] = ystart[i] + k3[i];
    }

    f(t + h, y, yin);           // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        k4[i] = h * yin[i];     // k4 = h * f(t+h, y(t)+k3)
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] = ystart[i] + k1[i] / 6 + k2[i] / 3 + k3[i] / 3 + k4[i] / 6;
    }
    // Return: y = new state at time t+h
}

